@@ -11,6 +11,7 @@ import itertools
 import logging
 from typing import Callable, Any, Dict, Iterable, NewType, Optional
 
+import async_timeout
 import objc
 from Foundation import NSNumber, NSObject, NSArray, NSData, NSError, NSUUID, NSString
 from CoreBluetooth import (
@@ -131,7 +132,10 @@ class PeripheralDelegate(NSObject):
 
     @objc.python_method
     async def read_characteristic(
-        self, characteristic: CBCharacteristic, use_cached: bool = True
+        self,
+        characteristic: CBCharacteristic,
+        use_cached: bool = True,
+        timeout: int = 20,
     ) -> NSData:
         if characteristic.value() is not None and use_cached:
             return characteristic.value()
@@ -141,7 +145,8 @@ class PeripheralDelegate(NSObject):
         self._characteristic_read_futures[characteristic.handle()] = future
         try:
             self.peripheral.readValueForCharacteristic_(characteristic)
-            return await asyncio.wait_for(future, timeout=5)
+            async with async_timeout.timeout(timeout):
+                return await future
         finally:
             del self._characteristic_read_futures[characteristic.handle()]
 
