@@ -2,10 +2,22 @@
 
 import asyncio
 import logging
-from typing import List
+import sys
+from typing import List, Optional
 import warnings
 
-from bleak.backends.scanner import BaseBleakScanner, AdvertisementData
+import async_timeout
+
+if sys.version_info[:2] < (3, 8):
+    from typing_extensions import Literal
+else:
+    from typing import Literal
+
+from bleak.backends.scanner import (
+    AdvertisementDataCallback,
+    BaseBleakScanner,
+    AdvertisementData,
+)
 from bleak.backends.device import BLEDevice
 from bleak.exc import BleakError
 
@@ -24,19 +36,32 @@ class BleakScannerP4Android(BaseBleakScanner):
     The python-for-android Bleak BLE Scanner.
 
     Args:
-        **detection_callback (callable or coroutine):
+        detection_callback:
             Optional function that will be called each time a device is
             discovered or advertising data has changed.
-        **service_uuids (List[str]):
+        service_uuids:
             Optional list of service UUIDs to filter on. Only advertisements
             containing this advertising data will be received. Specifying this
             also enables scanning while the screen is off on Android.
+        scanning_mode:
+            Set to ``"passive"`` to avoid the ``"active"`` scanning mode.
     """
 
     __scanner = None
 
-    def __init__(self, **kwargs):
-        super(BleakScannerP4Android, self).__init__(**kwargs)
+    def __init__(
+        self,
+        detection_callback: Optional[AdvertisementDataCallback] = None,
+        service_uuids: Optional[List[str]] = None,
+        scanning_mode: Literal["active", "passive"] = "active",
+        **kwargs,
+    ):
+        super(BleakScannerP4Android, self).__init__(detection_callback, service_uuids)
+
+        if scanning_mode == "passive":
+            self.__scan_mode = defs.ScanSettings.SCAN_MODE_OPPORTUNISTIC
+        else:
+            self.__scan_mode = defs.ScanSettings.SCAN_MODE_LOW_LATENCY
 
         self._devices = {}
         self.__adapter = None
@@ -52,7 +77,7 @@ class BleakScannerP4Android(BaseBleakScanner):
 
         logger.debug("Starting BTLE scan")
 
-        loop = asyncio.get_event_loop()
+        loop = asyncio.get_running_loop()
 
         if self.__javascanner is None:
             if self.__callback is None:
@@ -106,7 +131,7 @@ class BleakScannerP4Android(BaseBleakScanner):
             dispatchParams=(
                 filters,
                 defs.ScanSettingsBuilder()
-                .setScanMode(defs.ScanSettings.SCAN_MODE_LOW_LATENCY)
+                .setScanMode(self.__scan_mode)
                 .setReportDelay(0)
                 .setPhy(defs.ScanSettings.PHY_LE_ALL_SUPPORTED)
                 .setNumOfMatches(defs.ScanSettings.MATCH_NUM_MAX_ADVERTISEMENT)
@@ -121,7 +146,8 @@ class BleakScannerP4Android(BaseBleakScanner):
         self.__javascanner.flushPendingScanResults(self.__callback.java)
 
         try:
-            await asyncio.wait_for(scanfuture, timeout=0.2)
+            async with async_timeout.timeout(0.2):
+                await scanfuture
         except asyncio.exceptions.TimeoutError:
             pass
         except BleakError as bleakerror:
@@ -250,6 +276,11 @@ class _PythonScanCallback(utils.AsyncJavaCallbacks):
             entry.getKey().toString(): bytes(entry.getValue())
             for entry in record.getServiceData().entrySet()
         }
+        tx_power = result.getTxPower()
+
+        # change "not present" value to None to match other backends
+        if tx_power == defs.ScanResult.TX_POWER_NOT_PRESENT:
+            tx_power = None
 
         advertisement = AdvertisementData(
             local_name=record.getDeviceName(),
@@ -257,6 +288,7 @@ class _PythonScanCallback(utils.AsyncJavaCallbacks):
             service_data=service_data,
             service_uuids=service_uuids,
             platform_data=(result,),
+            tx_power=tx_power,
         )
         device = BLEDevice(
             device.getAddress(),
